/**
 * A single point in the source code input stream.
 */
export interface SourceLocation {
    /**
     * The character position in the stream (0-based).
     */
    offset: number;
    /**
     * The line number (1-based).
     */
    line: number;
    /**
     * The column number (1-based).
     */
    column: number;
    /**
     * The token index (0-based).
     */
    token?: number;
}
/**
 * The interval in the source code covered by a single
 * token or source code span.
 */
export interface SourceRange {
    /**
     * The beginning of the range (index of the first character).
     */
    start: SourceLocation;
    /**
     * The end of the range, immediately after the last character.
     */
    end: SourceLocation;
}
